# ltempy is a set of LTEM analysis and simulation tools developed by WSP as a member of the McMorran Lab
# Copyright (C) 2021  William S. Parker
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Contains utilities for reconstructing phase and magnetization from Lorentz images.

This module implements the Single Image Transport of Intensity Equation (SITIE) [1].

The most common use case is to generate a `SITIEImage` object from a `.dm3` or `.ser` file,
then reconstruct phase and induction with `sitie()`.

Example:

```python
import ncempy.io.dm as dm
import ltempy as lp

fname = '/path/to/data.dm3'
dm3file = dm.dmReader(fname)

img = lp.lorentz(dm3file)
img.reconstruct(defocus=1e-3)

### plot img.Bx, img.By, img.phase, img.data, etc
```

---

1. Chess, J. J. et al. Streamlined approach to mapping the magnetic induction of skyrmionic materials. Ultramicroscopy 177, 78–83 (2017).

"""

# %%
import numpy as np
import os

from . import process
from . import constants as _
np.seterr(divide='ignore', invalid='ignore')

__all__ = [
		'SITIEImage',
		'ind_from_phase',
		'sitie']

class SITIEImage:
	"""A SITIEImage object represents a defocussed image containing only magnetic contrast.

	**Parameters**

	* **datafile** : _dictionary_ <br />
	a dictionary with the following keys: <br />
		<ul>
		<li> **data** : _ndarray_ <br />
		A 2d array of the electron counts. </li>
		<li> **pixelSize** : _tuple_ <br />
		(_number_, _number_) - the x and y pixel sizes. </li>
		<li> **pixelUnit** : _tuple_ <br />
		(_string_, _string_) - the x and y pixel units. <br />
		Allowed values are `"pm"`, `"nm"`, `"µm"`, `"um"`, `"mm"`, and `"m"`.</li>
		</ul>
	"""
	def __init__(self, datafile):
		self.data = process.ndap(datafile.get('data')[::-1, :])
		self.dx = datafile.get('pixelSize')[0]
		self.dy = datafile.get('pixelSize')[1]
		self.x_unit = datafile.get('pixelUnit')[0]
		self.y_unit = datafile.get('pixelUnit')[1]
		self.x = np.arange(0,self.data.shape[1]) * self.dx
		self.y = np.arange(0,self.data.shape[0]) * self.dy
		self.phase = None
		self.Bx, self.By = None, None
		self.fix_units()

	def fix_units(self):
		"""Set the units to meters."""
		if self.x_unit == 'm':
			xr = 1
		elif self.x_unit == 'mm':
			xr = 1e-3
		elif self.x_unit == 'µm' or self.x_unit == 'um':
			xr = 1e-6
		elif self.x_unit == 'nm':
			xr = 1e-9
		elif self.x_unit == 'pm':
			xr = 1e-12
		else:
			print("Failed to set units to meters - pixelUnit was not recognized.")
			return(self)
		if self.y_unit == 'm':
			yr = 1
		elif self.y_unit == 'mm':
			yr = 1e-3
		elif self.y_unit == 'µm' or self.y_unit == 'um':
			yr = 1e-6
		elif self.y_unit == 'nm':
			yr = 1e-9
		elif self.y_unit == 'pm':
			yr = 1e-12
		else:
			print("Failed to set units to meters - pixelUnit was not recognized.")
			return(self)
		return(self.set_units(xr, "m", yr, "m"))

	def set_units(self, xr=1, x_unit="", yr=None, y_unit=None):
		"""Change the pixel units.

		**Parameters**

		* **xr** : _number, optional_ <br />
		New x-units per old x-units. (For example, 1e3 to convert from meters to mm). <br />
		Default is `xr = 1`.

		* **x_unit** : _string, optional_ <br />
		The new x-units (e.g., "nm" or "µm"). <br />
		Default is `x_unit = ""`.

		* **yr** : _number, optional_ <br />
		New y-units per old y-units. (For example, 1e3 to convert from meters to mm).
		If left empty, defaults to `xr`<br />
		Default is `yr = None`.

		* **y_unit** : _string, optional_ <br />
		The new y-units (e.g., "nm" or "µm"). If left empty, defaults to `x_unit`. <br />
		Default is `y_unit = None`.

		**Returns**

		* _SITIEImage_
		"""
		if yr is None:
			yr = xr
		if y_unit is None:
			y_unit = x_unit
		self.x_unit = x_unit
		self.y_unit = y_unit
		self.dx = self.dx * xr
		self.dy = self.dy * yr
		self.x = self.x * xr
		self.y = self.y * yr
		return(self)

	def reconstruct(self, df = 1e-3, thickness = 60e-9, wavelength=1.97e-12):
		"""Carries out phase and B-field reconstruction.

		Assigns `self.phase`, `self.Bx`, and `self.By` attributes.

		**Parameters**

		* **df** : _number, optional_ <br />
		The defocus at which the images were taken. <br />
		Default is `df = 1e-3`.

		* **thickness** : _number, optional_ <br />
		The thickness of the sample. <br />
		Default is `thickness = 60e-9`.

		* **wavelength** : _number, optional_ <br />
		The electron wavelength. <br />
		Default is `wavelength = 1.96e-12` (relativistic wavelength of a 300kV electron).

		**Returns**

		* _SITIEImage_
		"""
		self.phase = process.ndap(sitie(self.data, df, self.dx, self.dy, wavelength))
		self.Bx, self.By = [process.ndap(arr) for arr in ind_from_phase(self.phase, dx = self.dx, dy = self.dy, thickness = thickness)]
		return(self)

	def validate(self, threshold = 0.1, divangle = 1e-5, defocus = 1):
		r"""Estimate the validity of the SITIE approximation for the given parameters.

		SITIE expands on the approximations of TIE by assuming coherent illumination (that is,
		low divergence angle relative to the defocus and the spatial frequencies to be resolved).
		Another way to say this is that the damping envelope \(g(q_{\perp})\) is negligible.

		This returns \(q_{\perp}\) satisfying

		\[\frac{(\pi \Theta_c \Delta f)^2}{2} q_{\perp}^2 = \text{threshold}\]

		where \(\Delta f\) is the defocus and \(\Theta_c\) is the divergence angle. This \(q_{\perp}\)
		is the largest spatial frequency for which the SITIE approximation is valid.

		**Parameters**

		* **threshold** : _number, optional_ <br />
		The largest value that you consider \(<< 1\). <br />
		Default is `threshold = 0.1`.

		* **divangle** : _number, optional_ <br />
		The divergence angle. <br />
		Default is `divangle = 1e-5`.

		* **defocus** : _number, optional_ <br />
		Default is `defocus = 1`.

		* ****kwargs** <br />
		Any arguments to be passed to the damping function.
		"""
		prefactor = np.abs((_.pi * divangle * defocus)**2 / 2)
		qp_max = np.sqrt(threshold / prefactor)
		return(qp_max)

def sitie(img, defocus = 0, dx = 1, dy = 1, wavelength = 1.97e-12):
	"""Reconstruct the Aharonov-Bohm phase from a defocussed image.

	This is an implementation of the SITIE equation from [1], Eqn (10).

	**Parameters**

	* **img** : _ndarray_ <br />
	The 2d image data.

	* **defocus** : _number, optional_ <br />
	Default is `defocus = 0`.

	* **dx** : _number, optional_ <br />
	The pixel spacing in the x-direction. <br />
	Default is `dx = 1`

	* **dy** : _number, optional_ <br />
	The pixel spacing in the y-direction. <br />
	Default is `dy = 1`.

	* **wavelength** : _number, optional_ <br />
	The relativistic electron wavelength. <br />
	Default is `wavelength = 1.97e-9`.

	**Returns**

	* _ndarray_ <br />
	A 2d array containing the reconstructed Aharonov-Bohm phase shift.
	"""
	Sx = np.fft.fftfreq(img.shape[1], dx)
	Sy = np.fft.fftfreq(img.shape[0], dy)
	sx, sy = np.meshgrid(Sx, Sy)
	rhs = np.nan_to_num(2 * _.pi / wavelength / defocus * (1 - img / np.mean(img)), posinf = 0, neginf = 0)
	rhs = process.shift_pos(rhs)
	rhs = np.fft.fft2(rhs) / -4 / _.pi**2 / (sx**2 + sy**2)
	rhs = np.nan_to_num(rhs, posinf = 0, neginf = 0)
	phase = np.fft.ifft2(rhs)
	return(phase.real)

def ind_from_phase(phase, dx = 1, dy = 1, thickness = 60e-9):
	r"""Calculate the magnetic induction given the Aharonov-Bohm phase shift.

	Calculated using

	\[\nabla_{\perp} \phi_m = -\frac{e\tau}{\hbar} \left[\mathbf{B}\times\hat{\mathbf{e}}_{z}\right]\]

	where \(\tau\) is the thickness, \(\mathbf{B}\) is the transverse magnetic field components, integrated over \(z\),
	and \(\hat{\mathbf{e}}_z\) is the direction of propagation.

	**Parameters**

	* **phase** : _ndarray_ <br />
	A 2d array containing the electron phase immediately after the sample.

	* **dx** : _number, optional_ <br />
	The pixel spacing in the x-direction. <br />
	Default is `dx = 1`

	* **dy** : _number, optional_ <br />
	The pixel spacing in the y-direction. <br />
	Default is `dy = 1`.

	* **thickness** : _number, optional_ <br />
	The thickness of the sample. <br />
	Default is `thickness = 60e-9`.

	**Returns**

	* **Bx** : _ndarray_ <br />
	The x-component of the magnetic induction.

	* **By** : _ndarray_ <br />
	The y-component of the magnetic induction.
	"""
	dpdx = np.gradient(phase, dy, axis=1)
	dpdy = np.gradient(phase, dx, axis=0)
	Bx = _.hbar/_.e/thickness * dpdy
	By = -_.hbar/_.e/thickness * dpdx
	return(np.array([Bx.real, By.real]))



# ltempy is a set of LTEM analysis and simulation tools developed by WSP as a member of the McMorran Lab
# Copyright (C) 2021  William S. Parker
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""Allows unit scaling across `ltempy`.

By default, `ltempy` uses SI units. This module allows you to scale
the SI base units. So, for example, to work in kilometers rather than meters:

```python
import ltempy as lp
lp.set_units(meter = 1e-3) # set km as base unit for length

print(lp.constants.c) # yields 299792.458 km / s
```

Note that all other modules of `ltempy` and their functions update automatically as well.
So for example, if you set km as the base unit for length (`lp.set_units(meter = 1e-3)`),
and then run a sitie reconstruction with `defocus = 1`, you've just run sitie with defocus equal to 1km.

Also note that `ltempy` functions that use constants (such as `ind_from_phase`, which uses `constants.e` and
`constants.hbar`) will be affected, as both \(e\) (the electron charge) and \(\hbar\) (the reduced Planck constant),
have different numerical values when km is the base unit.
"""

__all__ = ['s','m','kg','A','K','mol','cd','c','hbar','e','pi','set_units']

def set_units(second=1, meter=1, kilogram=1, Ampere=1, Kelvin=1, mole=1, candela=1):
	"""Sets the units across the ltempy module.

	For example, `set_units(meter = 1e3)` sets the millimeter as the base unit for length.

	**Parameters**

	* **second** : _number, optional_ <br />
	The SI base unit for time. <br />
	Default is `second = 1`.

	* **meter** : _number, optional_ <br />
	The SI base unit for length. <br />
	Default is `meter = 1`.

	* **kilogram** : _number, optional_ <br />
	The SI base unit for mass. <br />
	Default is `kilogram = 1`.

	* **Ampere** : _number, optional_ <br />
	The SI base unit for current. <br />
	Default is `Ampere = 1`.

	* **Kelvin** : _number, optional_ <br />
	The SI base unit for temperature. <br />
	Default is `Kelvin = 1`.

	* **mole** : _number, optional_ <br />
	The SI base unit for amount of substance. <br />
	Default is `mole = 1`.

	* **candela** : _number, optional_ <br />
	The SI base unit for luminous intensity. <br />
	Default is `candela = 1`.
	"""
	global s,m,kg,A,K,mol,cd
	s,m,kg,A,K,mol,cd = second, meter, kilogram, Ampere, Kelvin, mole, candela
	set_consts(s, m, kg, A, K, mol, cd)

def set_consts(s, m, kg, A, K, mol, cd):
	"""Utility function for `set_units()`."""
	global c, hbar, e, pi
	F = s**4 * A**2 / m**2 / kg
	J = kg * m**2 / s**2
	C = A * s
	W = kg * m**2 / s**3
	eV = 1.602176634e-19 * J

	c       = 299792458.0   			* m / s
	hbar    = 1.0545718176461565e-34	* J * s
	e       = 1.602176634e-19			* C
	pi      = 3.141592653589793

set_units()



# ltempy is a set of LTEM analysis and simulation tools developed by WSP as a member of the McMorran Lab
# Copyright (C) 2021  William S. Parker
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Basic image processing tools.
"""

import os
from pathlib import Path
import numpy as np
from ._utils import _extend_and_fill_mirror
from warnings import warn

__all__ = ['high_pass', 'low_pass', 'gaussian_blur',
           'clip_data', 'shift_pos', 'outpath', 'multi', 'ndap']


def high_pass(data, cutoff=1 / 1024, dx=1, dy=1, padding=False):
    """Apply a high-pass filter to a 2d-array.

    **Parameters**

    * **data** : _complex ndarray_ <br />

    * **cutoff** : _number, optional_ <br />
    Cutoff frequency of the tophat filter
    or standard deviation of the gaussian filter, measured in inverse units. 
    If `dx` and `dy` are 1 (default), the unit is pixels. <br />
    Default is `cutoff = 1 / 1024`.

    * **dx** : _number, optional_ <br />
    Pixel spacing. <br />
    Default is `dx = 1`. 

    * **dy** : _number, optional_ <br />
    Pixel spacing. <br />
    Default is `dy = 1`. 

    * **padding** : _boolean, optional_ <br />
    Whether to zero-pad the input. Will use mirror padding if True. <br />
    Default is `padding = True`.

    **Returns**

    * _complex ndarray_ <br />
    """
    ds0, ds1 = data.shape[0], data.shape[1]
    if padding:
        bdata = _extend_and_fill_mirror(data)
    else:
        bdata = data
    X = np.fft.fftfreq(bdata.shape[1], dx)
    Y = np.fft.fftfreq(bdata.shape[0], dy)
    x, y = np.meshgrid(X, Y)

    g = np.zeros_like(x)

    g[x**2 + y**2 > cutoff**2] = 1
    g[x**2 + y**2 <= cutoff**2] = 0
    Fdata = np.fft.fft2(bdata)
    FFdata = np.fft.ifft2(g * Fdata)
    if padding:
        return(FFdata[ds0:2*ds0, ds1:2*ds1])
    else:
        return(FFdata)

def low_pass(data, cutoff=1 / 4, dx=1, dy=1, padding=False):
    """Apply a low-pass filter to a 2d-array.

    **Parameters**

    * **data** : _complex ndarray_ <br />

    * **cutoff** : _number, optional_ <br />
    Cutoff frequency of the tophat filter
    or standard deviation of the gaussian filter, measured in inverse units. 
    If `dx` and `dy` are 1 (default), the unit is pixels. <br />
    Default is `cutoff = 1 / 4`.

    * **dx** : _number, optional_ <br />
    Pixel spacing. <br />
    Default is `dx = 1`. 

    * **dy** : _number, optional_ <br />
    Pixel spacing. <br />
    Default is `dy = 1`. 

    * **padding** : _boolean, optional_ <br />
    Whether to zero-pad the input. Will use mirror padding if True. <br />
    Default is `padding = True`.

    **Returns**

    * _complex ndarray_ <br />
    """
    ds0, ds1 = data.shape[0], data.shape[1]
    if padding:
        bdata = _extend_and_fill_mirror(data)
    else:
        bdata = data
    X = np.fft.fftfreq(bdata.shape[1], dx)
    Y = np.fft.fftfreq(bdata.shape[0], dy)
    x, y = np.meshgrid(X, Y)

    g = np.zeros_like(x)

    g[x**2 + y**2 < cutoff**2] = 1
    g[x**2 + y**2 >= cutoff**2] = 0
    Fdata = np.fft.fft2(bdata)
    FFdata = np.fft.ifft2(g * Fdata)
    if padding:
        return(FFdata[ds0:2*ds0, ds1:2*ds1])
    else:
        return(FFdata)

def gaussian_blur(data, blur_radius=1, padding = True):
    """Apply a Gaussian blur to the data. 

    **Parameters**

    * **data** : _complex ndarray_ <br />

    * **blur_radius** : _number, optional_ <br />
    The standard deviation of the Gaussian kernel, measured in pixels. <br />
    Default is `blur_radius = 1`.

    * **padding** : _boolean, optional_ <br />
    Whether to zero-pad the input. Will use mirror padding if True. <br />
    Default is `padding = True`.

    **Returns**

    * _complex ndarray_ <br />
    """
    ds0, ds1 = data.shape[0], data.shape[1]
    if padding:
        bdata = _extend_and_fill_mirror(data)
    else:
        bdata = data
    X = np.fft.fftfreq(bdata.shape[1], 1)
    Y = np.fft.fftfreq(bdata.shape[0], 1)
    x, y = np.meshgrid(X, Y)

    g = np.exp(-2 * np.pi**2 * blur_radius**2 * (x**2 + y**2))

    Fdata = np.fft.fft2(bdata)
    FFdata = np.fft.ifft2(g * Fdata)
    if padding:
        return(FFdata[ds0:2*ds0, ds1:2*ds1])
    else:
        return(FFdata)


def multi(data, opts, dx=1, dy=1):
    """Perform multiple filters at once, but only performs one Fourier / inverse Fourier transform. 
    
    **Parameters**

    * **data** : _complex ndarray_ <br />

    * **opts** : _dictionary_ <br />
    A dictionary containing the processing to perform. Format is 
    ```
    {
      'gaussian_blur': {'blur_radius': n},
      'high_pass': {'cutoff': m},
      'low_pass': {'cutoff': l}
    }
    ```
    Include a filter simply by including it as a key. If a filter is included, 
    its parameters must match the format above. 

    * **dx** : _float, optional_ <br />
    Pixel spacing. <br />
    Default is `dx = 1`.

    * **dy** : _float, optional_ <br />
    Pixel spacing. <br />
    Default is `dy = 1`.

    **Returns**

    * _complex ndarray_ <br />
    
    """
    ds0, ds1 = data.shape[0], data.shape[1]
    bdata = _extend_and_fill_mirror(data)
    X = np.fft.fftfreq(bdata.shape[1], dx)
    Y = np.fft.fftfreq(bdata.shape[0], dy)
    x, y = np.meshgrid(X, Y)
    Fdata = np.fft.fft2(bdata)
    g = np.ones_like(Fdata)
    if "gaussian_blur" in opts:
      g = g * np.exp(-2 * np.pi**2 * opts['gaussian_blur']['blur_radius']**2 * (x**2 + y**2))
    if "high_pass" in opts:
      g[x**2 + y**2 <= opts['high_pass']['cutoff']**2] = 0
    if "low_pass" in opts:
      g[x**2 + y**2 >= opts['low_pass']['cutoff']**2] = 0
    FFdata = np.fft.ifft2(g * Fdata)
    return(FFdata[ds0:2*ds0, ds1:2*ds1])

        


def clip_data(data, sigma=5):
    """Clip data to a certain number of standard deviations from its average.

    **Parameters**

    * **data** : _complex ndarray_ <br />

    * **sigma** : _number, optional_ <br />
    Number of standard deviations from average to clip to. <br />
    Default is `sigma = 5`.

    **Returns**

    * _complex ndarray_ <br />
    """
    avg = np.mean(data)
    stdev = np.std(data)
    vmin = avg - sigma*stdev
    vmax = avg + sigma*stdev
    out = data.copy()
    out[out < vmin] = vmin
    out[out > vmax] = vmax
    return(out)

def shift_pos(data):
    """Shift data to be positive.

    **Parameters**

    * **data** : _ndarray_ <br />
    Dtype must be real.

    **Returns**

    * _ndarray_
    """
    return(data - np.min(data))


def outpath(datadir, outdir, fname, create=False):
    """A util to get the output filename.

    An example is easiest to explain:

    datadir: `/abs/path/to/data`

    fname: `/abs/path/to/data/plus/some/structure/too.dm3`

    outdir: `/where/i/want/to/write/data`

    This util can create the folder (if it doesn't exist):

    `/where/i/want/to/write/data/plus/some/structure`

    and return the filename:

    `/where/i/want/to/write/data/plus/some/structure/too`.

    **Parameters**

    * **datadir** : _string_ <br />
    The directory for the experiment's data.

    * **outdir** : _string_ <br />
    The main directory where you want outputs to go.

    * **fname** : _string_ <br />
    The name of the file in datadir.

    * **create** : _boolean, optional_ <br />
    Whether to create the output directory. <br />
    Default is `create = False`.

    **Returns**

    * _PurePath_ <br />
    The name of the file to save (without a suffix).
    """
    fname = os.path.splitext(fname)[0]
    subpath = os.path.relpath(os.path.dirname(fname), datadir)
    finoutdir = os.path.join(outdir, subpath)
    if create:
        if not os.path.exists(finoutdir):
            os.makedirs(finoutdir)
    return(Path(os.path.join(finoutdir, os.path.basename(fname))))


class ndap(np.ndarray):
    """A class that adds all the image processing methods to `np.ndarray`.

    The purpose of this class is just so you can write `myarray.high_pass().low_pass()`
    instead of `myarray = high_pass(low_pass(myarray))`.

    If `x` and `y` are given, then `dx` and `dy` are calculated and stored. 
    They will default to `dx = 1` and `dy = 1`, and will be used in `low_pass()`, 
    `high_pass()`, etc, unless specified otherwise in the function call. 

    **Parameters**

    * **data** : _complex ndarray_ <br />
    Any type of ndarray - the methods are defined with a 2d array in mind.

    * **x** : _ndarray_ <br />
    The x coordinates associated with the array data. 
    Must be a 1-dimensional ndarray with `len(x) == data.shape[1]`. 
    Note that `x` is associated with the second axis, to match the 
    convention of `meshgrid`.

    * **y** : _ndarray_ <br />
    The y coordinates associated with the array data. 
    Must be a 1-dimensional ndarray with 'len(y) == data.shape[0]`. 
    Note that `y` is associated with the first axis, to match the 
    convention of `meshgrid`.

    """
    ## __new__ and __array_finalize__ are directly from numpy docs
    ## their example is almost identical to this
    ## subclassing numpy arrays
    def __new__(cls, data, x=None, y=None):
        obj = np.asarray(data, dtype=data.dtype).copy().view(cls)
        if x is None:
          obj.x = np.arange(data.shape[1])
          obj.dx = obj.x[1] - obj.x[0]
        else:
          try: sel = obj.shape[1] != x.shape[0] or len(x.shape) != 1
          except: raise Exception("x must be a 1-dimensional numpy.ndarray. ")
          if sel:
            raise Exception("x shape must match the data's 1st axis. ")
          obj.x = x
          obj.dx = x[1] - x[0]
        if y is None:
          obj.y = np.arange(data.shape[0])
          obj.dy = obj.y[1] - obj.y[0]
        else:
          try: sel = obj.shape[0] != y.shape[0] or len(y.shape) != 1
          except: raise Exception("y must be a 1-dimensional numpy.ndarray. ")
          if sel:
            raise Exception("y shape must match the data's 1st axis. ")
          obj.y = y
          obj.dy = y[1] - y[0]
        obj.isComplex = np.iscomplexobj(data)
        return obj

    def __array_finalize__(self, obj):
      if obj is None: return
      self.x = getattr(obj, 'x', None)
      self.y = getattr(obj, 'y', None)
      self.dx = getattr(obj, 'dx', None)
      self.dy = getattr(obj, 'dy', None)
      self.isComplex = getattr(obj, 'isComplex', None)

    def high_pass(self, cutoff=1 / 1024, dx=None, dy=None, padding=False):
        """Apply a high-pass filter to a 2d-array.

        **Parameters**

        * **cutoff** : _number, optional_ <br />
        Cutoff frequency of the tophat filter
        or standard deviation of the gaussian filter, measured in inverse units. 
        If `dx` and `dy` are 1 (default), the unit is pixels. <br />
        Default is `cutoff = 1 / 1024`.

        * **dx** : _number, optional_ <br />
        Pixel spacing. <br />
        Default is `dx = self.dx`. 

        * **dy** : _number, optional_ <br />
        Pixel spacing. <br />
        Default is `dy = self.dy`. 

        * **padding** : _boolean, optional_ <br />
        Whether to zero-pad the input. Uses mirror padding if True. <br />
        Default is `padding = False`.

        **Returns**

        * _ndap_ <br />
        """
        if dx is None:
          dx = self.dx
        if dy is None:
          dy = self.dy
        if self.isComplex:
            self[:, :] = high_pass(self, cutoff, dx, dy, padding)
        else:
            self[:, :] = np.real(high_pass(self, cutoff, dx, dy, padding))
        return(self)

    def low_pass(self, cutoff=1 / 4, dx=None, dy=None, padding=False):
        """Apply a low-pass filter to a 2d-array.

        **Parameters**

        * **cutoff** : _number, optional_ <br />
        Cutoff frequency of the tophat filter
        or standard deviation of the gaussian filter, measured in inverse units. 
        If `dx` and `dy` are 1 (default), the unit is pixels. <br />
        Default is `cutoff = 1 / 4`.

        * **dx** : _number, optional_ <br />
        Pixel spacing. <br />
        Default is `dx = self.dx`. 

        * **dy** : _number, optional_ <br />
        Pixel spacing. <br />
        Default is `dy = self.dy`. 

        * **padding** : _boolean, optional_ <br />
        Whether to zero-pad the input. Uses mirror padding if True. <br />
        Default is `padding = False`.

        **Returns**

        * _ndap_ <br />
        """
        if dx is None:
          dx = self.dx
        if dy is None:
          dy = self.dy
        if self.isComplex:
            self[:, :] = low_pass(self, cutoff, dx, dy, padding)
        else:
            self[:, :] = np.real(low_pass(self, cutoff, dx, dy, padding))
        return(self)
    

    def multi(self, opts, dx=None, dy=None):
        """Perform multiple filters at once. Only performs one Fourier / inverse Fourier transform. 
        
        **Parameters**

        * **data** : _complex ndarray_ <br />

        * **opts** : _dictionary_ <br />
        A dictionary containing the processing to perform. Format is 
        ```
        {
          'gaussian_blur': {'blur_radius': n},
          'high_pass': {'cutoff': m},
          'low_pass': {'cutoff': l}
        }
        ```
        . Include a filter simply by including it as a key. If a filter is included, 
        its parameters must match the format above. 

        * **dx** : _float, optional_ <br />
        Pixel spacing. <br />
        Default is `dx = self.dx`.

        * **dy** : _float, optional_ <br />
        Pixel spacing. <br />
        Default is `dy = self.dy`.

        **Returns**

        * _complex ndarray_ <br />
        """ 
        if dx is None:
          dx = self.dx
        if dy is None:
          dy = self.dy
        if self.isComplex:
          self[:, :] = multi(self, opts, dx, dy)
        else:
          self[:, :] = np.real(multi(self, opts, dx, dy))
        return(self)

    def gaussian_blur(self, blur_radius=1, padding=True):
        """Apply a Gaussian blur to the data. 

        **Parameters**

        * **data** : _complex ndarray_ <br />

        * **blur_radius** : _number, optional_ <br />
        The standard deviation of the Gaussian kernel, measured in pixels. <br />
        Default is `blur_radius = 1`.

        * **padding** : _boolean, optional_ <br />
        Whether to zero-pad the input. Uses mirror padding if True. <br />
        Default is `padding = False`.

        **Returns**

        * _ndap_ <br />
        """
        if self.isComplex:
            self[:, :] = gaussian_blur(self, blur_radius, padding)
        else:
            self[:, :] = np.real(gaussian_blur(self, blur_radius, padding))
        return(self)
      
    def get_window(self, window, step=1):
      """Get a windowed section of data. 

      **Parameters**

      * **window** : _tuple_ <br />
      [xmin, xmax, ymin, ymax]. Note that these are the x and y values, not their indices. 
      Also note that the default x and y attributes of the ndap object are just its indices. 

      **Returns**

      * _ndap_ <br />
      A 2d ndap with the windowed data and new axes.
      """
      argxmin = np.argmin(np.abs(self.x - window[0]))
      argxmax = np.argmin(np.abs(self.x - window[1]))
      argymin = np.argmin(np.abs(self.y - window[2]))
      argymax = np.argmin(np.abs(self.y - window[3]))
      xout = self.x[argxmin:argxmax:step].copy()
      yout = self.y[argymin:argymax:step].copy()
      dout = ndap(self[argymin:argymax:step, argxmin:argxmax:step].copy(), x=xout, y=yout)
      return(dout)

    def clip_data(self, sigma=5):
        """Clip data to a certain number of standard deviations from its average.

        **Parameters**

        * **sigma** : _number, optional_ <br />
        Number of standard deviations from average to clip to. <br />
        Default is `sigma = 5`.

        **Returns**

        * _ndap_ <br />
        """
        self[:, :] = clip_data(self, sigma)
        return(self)

    def shift_pos(self):
        """Shift data to be positive.

        **Returns**

        * _ndap_ <br />
        """
        if self.isComplex:
            print("Positive shift not applied to complex data.")
            return(self)
        self[:, :] = shift_pos(self)
        return(self)



# ltempy is a set of LTEM analysis and simulation tools developed by WSP as a member of the McMorran Lab
# Copyright (C) 2021  William S. Parker
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Generate rgba data from scalar values and implement the CIELAB colorspace.

The CIELAB colorspace is intended to be perceptually uniform. While it doesn't fully succeed,
it does well enough to use color and brightness as independent information channels.

In this module, the CIELAB colorspace is used as a representation of complex values,
where color represents phase and brightness represents magnitude.
"""

import numpy as np
from matplotlib.colors import ListedColormap
import matplotlib.pyplot as plt
from . import constants as _

__all__ = ['cielab_rgba',
            'rgba',
            'cielab_cmap']

def cielab_cmap(samples=256):
    """Creates a `matplotlib.colors.ListedColormap` of the CIELAB color space.

    **Parameters**

    * **samples** : _integer, optional_ <br />
    The number of samples. Any additional values will be nearest-neighbor interpolated by matplotlib. <br />
    Default is `samples = 256`.

    **Returns**

    * _ListedColormap_ <br />
    A colormap. Many `matplotlib` functions that accept strings for the cmap argument
    also accept a `ListedColormap`: `plt.imshow(data, cmap=cmap)`.
    """
    angles = np.linspace(0,2*_.pi,samples)
    cvals = np.exp(1j*angles).reshape(1, samples)
    rgbavals = cielab_rgba(cvals).squeeze()/255
    cmap = ListedColormap(rgbavals)
    return(cmap)

def cielab_rgba(data, brightness = 'intensity', alpha = 'uniform', shift=0):
    """Convert complex values to rgba values based on the CIELAB color space.

    Color represents the complex phase, while brightness may represent amplitude,
    intensity (amplituded squared) or be uniform.

    The CIELAB color space is designed to be perceptually uniform - none of the
    colors appear brighter or darker than the others.

    **Parameters**

    * **data** : _ndarray_ <br />
    An array with the data to convert. Dtype may be complex or real - if real,
    the color will be uniform, and values will be represented by brightness. Does not need to be normalized.

    * **brightness** : _string, optional_ <br />
    Allowed values: `'intensity'`, `'amplitude'`, `'uniform'`. <br />
    Default is `brightness = 'intensity'`.

    * **alpha** : _string, optional_ <br />
    Allowed values: `'intensity'`, `'amplitude'`, `'uniform'`. Determines the alpha
    component of the rgba value. <br />
    Default is `alpha = 'uniform'`.

    * **shift** : _number, optional_ <br />
    Angle by which to shift the colors. Does not modify the angle of the data. <br />
    Default is `shift = 0`.  

    **Returns**

    * _ndarray_ <br />
    The calculated rgba components. If the input array has
    shape MxN, the output array will have shape MxNx4.
    """
    vals = {
        'uniform': lambda data: 255,
        'amplitude': lambda data: np.absolute(data) * 255,
        'intensity': lambda data: np.absolute(data)**2 * 255
        }
    if not brightness in vals.keys():
        raise ValueError("Invalid brightness value; should be one of {}".format(list(vals.keys())))
    if not alpha in vals.keys():
        raise ValueError("Invalid alpha value; should be one of {}".format(list(vals.keys())))

    data = data / np.max(np.abs(data))
    rgba_image_components = np.zeros(np.append(data.shape, 4),dtype=np.uint8)
    bvalue = vals[brightness](data)
    avalue = vals[alpha](data)
    hue = (np.angle(data) + np.pi - shift) / 2
    pi6 = np.pi/6
    def sin2(array, offset):
        return(np.sin(array-offset)**2)
    r = sin2(hue, .15*pi6) + 0.35 * sin2(hue, 3.15 * pi6)
    b = sin2(hue, 4.25 * pi6)
    g = .6*sin2(hue, 2*pi6) + 0.065 * sin2(hue, 5.05 * pi6) + 0.445*b + 0.33*r
    rgba_image_components[...,0] = (r * bvalue).astype(np.uint8)
    rgba_image_components[...,1] = (g * bvalue).astype(np.uint8)
    rgba_image_components[...,2] = (b * bvalue).astype(np.uint8)
    rgba_image_components[...,3] = np.full(data.shape, fill_value=avalue, dtype=np.uint8)
    return(rgba_image_components)

def rgba(mode, cmap = None, brightness = 'intensity', alpha = 'uniform', shift=0):
    """Converts a 2d complex array to rgba data.

    **Parameters**

    * **mode** : _complex ndarray_ <br />
    An array with the data to represent. Dtype may be complex or real.

    * **cmap** : _string, optional_ <br />
    A `matplotlib` ScalarMappable to use for colors. If left blank,
    the CIELAB color space will be used instead. <br />
    Default is `cmap = None`.

    * **brightness** : _string, optional_ <br />
    Allowed values: `'intensity'`, `'amplitude'`, `'uniform'`. <br />
    Default is `brightness = 'intensity'`.

    * **alpha** : _string, optional_ <br />
    Allowed values: `'intensity'`, `'amplitude'`, `'uniform'`. Determines the alpha (opacity)
    component of the rgba value. <br />
    Default is `alpha = 'uniform'`.

    * **shift** : _number, optional_ <br />
    Angle by which to shift the colors. Does not modify the angle of the data. <br />
    Default is `shift = 0`.  

    **Returns**

    * _ndarray_ <br />
    The calculated rgba components. If the input array has
    shape NxN, the output array will have shape NxNx4.
    """
    vals = {
        'uniform': lambda data: 255,
        'amplitude': lambda data: 255 * np.absolute(data),
        'intensity': lambda data: 255 * np.absolute(data)**2
        }
    if not brightness in vals.keys():
        raise ValueError("Invalid brightness value; should be one of {}".format(list(vals.keys())))
    if not alpha in vals.keys():
        raise ValueError("Invalid alpha value; should be one of {}".format(list(vals.keys())))
    if cmap is None:
        return(cielab_rgba(mode, brightness, alpha, shift))
    mode = mode / np.max(np.absolute(mode))
    colormap = plt.cm.ScalarMappable(cmap=cmap)
    out = colormap.to_rgba((np.angle(mode) - shift)%(2 * np.pi))
    out[...,-1] = vals[alpha](mode)
    out[...,0] = out[...,0] * vals[brightness](mode)
    out[...,1] = out[...,1] * vals[brightness](mode)
    out[...,2] = out[...,2] * vals[brightness](mode)
    return(out.astype(np.uint8))



# ltempy is a set of LTEM analysis and simulation tools developed by WSP as a member of the McMorran Lab
# Copyright (C) 2021  William S. Parker
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
r"""Simulate Aharonov-Bohm phase, magnetic vector potential, magnetic field, and LTEM images.

Given a magnetic thin film with magnetization \(\mathbf{m}(x,y)\) and thickness \(\tau\), one can
come to analytical expressions [1] for the Fourier components of

1. the magnetic vector potential \(\mathbf{A}(x, y, z)\)
2. the magnetic field \(\mathbf{B}(x, y, z)\)
3. the Aharonov-Bohm phase \(\phi_m(x, y)\) acquired by an electron passing through the sample

From there, the magnetic vector potential, magnetic field, and phase can be calculated directly
using a fast fourier transform algorithm.

The Aharonov-Bohm phase can be related to the integrated perpendicular components of magnetic field via

\[
\nabla_{\perp} \phi_m = -\frac{e\tau}{\hbar} \left[\mathbf{B}\times\hat{\mathbf{e}}_{z}\right]
\]

where \(\hat{\mathbf{e}}_z\) is the direction of propagation.

Lorentz TEM images can be calculated three ways [2] (all within the framework of Fourier optics):

1. `img_from_mag`: Directly from the magnetization, using the paraxial approximation.
2. `img_from_phase`: From a phase, using the paraxial approximation.
3. `propagate`: From an exit wavefunction, without the paraxial approximation.

---

1. Mansuripur, M. Computation of electron diffraction patterns in Lorentz electron microscopy of thin magnetic films. Journal of Applied Physics 69, 2455–2464 (1991).

2. Chess, J. J. et al. Streamlined approach to mapping the magnetic induction of skyrmionic materials. Ultramicroscopy 177, 78–83 (2017).

"""

from . import constants as _
import numpy as np
from .process import shift_pos
from .sitie import ind_from_phase
from ._utils import T, sims_shared, weights, G, A_mn_components, laplacian_2d, inverse_laplacian_2d, gradient_2d

__all__ = [
			'ab_phase',
			'B_from_mag',
			'A_from_mag',
			'img_from_mag',
			'img_from_phase',
			'jchessmodel',
			'ind_from_mag',
			'propagate']

# Mansuripur
def ab_phase(mx, my, mz, dx=1, dy=1, thickness=60e-9, p = np.array([0,0,1])):
	"""Calculate the Aharonov-Bohm phase imparted on a fast electron by a magnetized sample.

	This is a direct implementation of [1], Eq (13).

	The shape of the output array is the same as mx, my, and mz. If mx, my, mz are three dimensional
	(i.e., have z-dependence as well as x and y), the third dimension of the output array represents
	the Aharonov-Bohm phase of each z-slice of the magnetization.

	**Parameters**

	* **mx** : _ndarray_ <br />
	The x-component of magnetization. Should be two or three dimensions.

	* **my** : _ndarray_ <br />
	The y-component of magnetization. Should be two or three dimensions.

	* **mz** : _ndarray_ <br />
	The z-component of magnetization. Should be two or three dimensions.

	* **dx** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the x-direction. <br />
	Default is `dx = 1`.

	* **dy** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the y-direction. <br />
	Default is `dy = 1`.

	* **thickness** : _number, optional_ <br />
	The thickness of each slice of the x-y plane (if no z-dependence, the thickness of the sample). <br />
	Default is `thickness = 60e-9`.

	* **p** : _ndarray, optional_ <br />
	A unit vector representing the direction of the electron's path. Shape should be `(3,)`. <br />
	Default is `p = np.array([0,0,1])`.

	**Returns**

	* _ndarray_ <br />
	The Aharonov-Bohm phase imparted on the electron by the magnetization. Shape will be the same as mx, my, mz.
	"""
	# All direct from definitions in Mansuripur
	M, s, s_mag, sig, z_hat = sims_shared(mx, my, mz, dx, dy)
	w = weights(M, s, s_mag, sig, z_hat, p, thickness)

	# multiply by weights.shape to counter ifft2's normalization
	# old versions of numpy don't have the `norm = 'backward'` option
	phase = w.shape[1] * w.shape[2] * np.fft.ifft2(w, axes=(1,2))
	return(np.squeeze(phase.real))

def B_from_mag(mx, my, mz, dx = 1, dy = 1, z = 0, thickness = 60e-9):
	r"""Calculate the magnetic field of a specified 2-d magnetic configuration.

	This is an implementation of [1], Eqn (11), which gives an analytic expression for
	the Fourier components of the magnetic vector potential. The magnetic field Fourier components
	are then calculated analytically via \(\mathbf{B} = \nabla\times\mathbf{A}\).

	The output shape is `(3, mx.shape[0], mx.shape[1], z.shape[0])`.

	**Parameters**

	* **mx** : _ndarray_ <br />
	The x-component of magnetization. Must be a 2-d array.

	* **my** : _ndarray_ <br />
	The y-component of magnetization. Must be a 2-d array.

	* **mz** : _ndarray_ <br />
	The z-component of magnetization. Must be a 2-d array.

	* **dx** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the x-direction. <br />
	Default is `dx = 1`.

	* **dy** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the y-direction. <br />
	Default is `dy = 1`.

	* **z** : _number, optional_, ndarray_ <br />
	The z-coordinates at which to calculate the B-field. Can be a number, or a 1d-array. <br />
	Default is `z = 0`.

	* **thickness** : _number, optional_ <br />
	The thickness of the sample. <br />
	Default is `thickness = 60e-9`.

	**Returns**

	* _ndarray_ <br />
	The magnetic field resulting from the given 2d magnetization.
	"""
	z = np.atleast_1d(z)
	selz_m = z < - thickness / 2
	selz_z = np.abs(z) <= thickness / 2
	selz_p = z > thickness / 2
	z = z[np.newaxis,np.newaxis,np.newaxis,...]

	M, s, s_mag, sig, z_hat = sims_shared(mx, my, mz, dx, dy)
	sigp = sig + 1j * z_hat
	sigm = sig - 1j * z_hat

	A_mn = A_mn_components(
			mx.shape[0], mx.shape[1], z.shape[-1], selz_m, selz_z,
			selz_p, s_mag, z, sigm, sigp, sig, M, thickness, z_hat)
	dxA_mn = 1j * 2 * _.pi * s[0][np.newaxis, ...] * A_mn
	dyA_mn = 1j * 2 * _.pi * s[1][np.newaxis, ...] * A_mn
	dzA_mn = np.zeros((3, mx.shape[0], mx.shape[1], z.shape[-1]), dtype=complex)
	dzA_mn[...,selz_m] = 2 * _.pi * s_mag * A_mn[...,selz_m]
	dzA_mn[...,selz_p] = -2 * _.pi * s_mag * A_mn[...,selz_p]
	dzA_mn[...,selz_z] = (2 * 1j / s_mag * np.cross((
						sig
						- 0.5 * 2 * _.pi * s_mag * np.exp(2 * _.pi * s_mag * (z[...,selz_z] - thickness / 2)) * sigm
						+ 0.5 * 2 * _.pi * s_mag * np.exp(-2 * _.pi * s_mag * (z[...,selz_z] + thickness / 2)) * sigp
						), M, axisa=0, axisb=0, axisc=0))
	dzA_mn[:,0,0,selz_m] = 0
	dzA_mn[:,0,0,selz_p] = 0
	dzA_mn[:,0,0,selz_z] = -4 * _.pi * np.cross(z_hat[:,0,0,:], M[:,0,0,:], axisa=0, axisb=0, axisc=0)
	B_mn = np.zeros((3, mx.shape[0], mx.shape[1], z.shape[-1]), dtype=complex)
	B_mn[0] = dyA_mn[2] - dzA_mn[1]
	B_mn[1] = dzA_mn[0] - dxA_mn[2]
	B_mn[2] = dxA_mn[1] - dyA_mn[0]
	B = B_mn.shape[1] * B_mn.shape[2] * np.fft.ifft2(B_mn, axes=(1,2))
	return(np.squeeze(B.real))

def A_from_mag(mx, my, mz, dx = 1, dy = 1, z = 0, thickness = 60e-9):
	r"""Calculate the magnetic vector potential of a specified 2-d magnetic configuration.

	This is an implementation of [1], Eqn (11), which gives the Fourier components of
	the magnetic vector potential.

	The output shape is `(3, mx.shape[0], mx.shape[1], z.shape[0])`.

	**Parameters**

	* **mx** : _ndarray_ <br />
	The x-component of magnetization. Must be a 2-d array.

	* **my** : _ndarray_ <br />
	The y-component of magnetization. Must be a 2-d array.

	* **mz** : _ndarray_ <br />
	The z-component of magnetization. Must be a 2-d array.

	* **dx** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the x-direction. <br />
	Default is `dx = 1`.

	* **dy** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the y-direction. <br />
	Default is `dy = 1`.

	* **z** : _number, optional_, ndarray_ <br />
	The z-coordinates at which to calculate the B-field. Can be a number_, or a 1d-array. <br />
	Default is `z = 0`.

	* **thickness** : _number, optional_ <br />
	The thickness of the sample. <br />
	Default is `thickness = 60e-9`.

	**Returns**

	* _ndarray_ <br />
	The magnetic vector potential resulting from the given 2d magnetization.
	"""
	z = np.atleast_1d(z)
	selz_m = z < -thickness/2
	selz_z = np.abs(z) <= thickness/2
	selz_p = z > thickness/2
	z = z[np.newaxis,np.newaxis,np.newaxis,...]

	M, s, s_mag, sig, z_hat = sims_shared(mx, my, mz, dx, dy)
	sigp = sig + 1j * z_hat
	sigm = sig - 1j * z_hat

	A_mn = A_mn_components(
				mx.shape[0], mx.shape[1], z.shape[-1], selz_m, selz_z,
				selz_p, s_mag, z, sigm, sigp, sig, M, thickness, z_hat)
	A = A_mn.shape[1] * A_mn.shape[2] * np.fft.ifft2(A_mn, axes=(1,2))
	return(np.squeeze(A.real))

def ind_from_mag(mx, my, mz, dx=1, dy=1, thickness=60e-9, p = np.array([0,0,1])):
	"""Calculate the integrated perpendicular magnetic field of a magnetized sample.

	This is shorthand for `ind_from_phase(ab_phase(*args))`. This method is used rather than `B_from_mag`
	because the integral over z can be done analytically.

	**Parameters**

	* **mx** : _ndarray_ <br />
	The x-component of magnetization. Should be two or three dimensions.

	* **my** : _ndarray_ <br />
	The y-component of magnetization. Should be two or three dimensions.

	* **mz** : _ndarray_ <br />
	The z-component of magnetization. Should be two or three dimensions.

	* **dx** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the x-direction. <br />
	Default is `dx = 1`.

	* **dy** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the y-direction. <br />
	Default is `dy = 1`.

	* **thickness** : _number, optional_ <br />
	The thickness of each slice of the x-y plane (if no z-dependence, the thickness of the sample). <br />
	Default is `thickness = 60e-9`.

	* **p** : _ndarray, optional_ <br />
	A unit vector representing the direction of the electron's path. Shape should be `(3,)`. <br />
	Default is `p = np.array([0,0,1])`.

	**Returns**

	* _ndarray_ <br />
	The x-component of the magnetic induction.

	* _ndarray_ <br />
	The y-component of the magnetic induction.
	"""
	phase = ab_phase(mx, my, mz, dx, dy, thickness, p)
	Bx, By = ind_from_phase(phase, dx, dy, thickness)
	return(np.array([Bx, By]))

def img_from_mag(mx, my, mz, dx = 1, dy = 1, defocus = 0, thickness = 60e-9, wavelength = 1.97e-12, p = np.array([0,0,1]), divangle = 1e-5):
	r"""Calculate the Lorentz TEM image from a given (2 or 3 dim) magnetization.

	This is a combination of [2], Eqn (7), which gives the output intensity in terms of \(\phi_m\) within the paraxial approximation,
	and [1], Eqn (13), which gives the Fourier components of \(\phi_m\) in terms of the magnetization.

	**Parameters**

	* **mx** : _ndarray_ <br />
	The x-component of magnetization. Should be two or three dimensions.

	* **my** : _ndarray_ <br />
	The y-component of magnetization. Should be two or three dimensions.

	* **mz** : _ndarray_ <br />
	The z-component of magnetization. Should be two or three dimensions.

	* **dx** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the x-direction. <br />
	Default is `dx = 1`.

	* **dy** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the y-direction. <br />
	Default is `dx = 1`.

	* **defocus** : _number, optional_ <br />
	The defocus - note that this should be non-zero in order to see any contrast. <br />
	Default is `defocus = 0`.

	* **thickness** : _number, optional_ <br />
	The thickness of each slice of the x-y plane (if no z-dependence, the thickness of the sample). <br />
	Default is `thickness = 60e-9`.

	* **wavelength** : _number, optional_ <br />
	The relativistic electron wavelength. <br />
	Default is `wavelength = 1.97e-12`.

	* **p** : _ndarray, optional_ <br />
	A unit vector representing the direction of the electron's path. Shape should be `(3,)`. <br />
	Default is `p = np.array([0,0,1])`.

	* **divangle** : _number, optional_ <br />
	The divergence angle \(\Theta_c\). <br />
	Default is `divangle = 1e-5`.

	**Returns**

	* _ndarray_ <br />
	The intensity of the image plane.
	"""
	M, s, s_mag, sig, z_hat = sims_shared(mx, my, mz, dx, dy)
	w = weights(M, s, s_mag, sig, z_hat, p, thickness)

	nabla2weights = - 4 * _.pi**2 * s_mag**2 * w
	nablaweights = 1j * 2 * _.pi * s * w

	nabla2phi = nabla2weights.shape[1] * nabla2weights.shape[2] * np.fft.ifft2(nabla2weights, axes=(1,2))
	nablaphi = nablaweights.shape[1] * nablaweights.shape[2] * np.fft.ifft2(nablaweights, axes=(1,2))
	nablaphi2 = nablaphi[0]**2 + nablaphi[1]**2

	if nablaphi2.shape[-1] > 1:
		nablaphi2 = np.sum(nablaphi2, axis=-1)
	if nabla2phi.shape[-1] > 1:
		nabla2phi = np.sum(nabla2phi, axis=-1)
	out = 1 - wavelength * defocus / 2 / _.pi * np.squeeze(nabla2phi) - (_.pi * divangle * defocus)**2 / 2 / np.log(2) * np.squeeze(nablaphi2)
	return(out.real)

def img_from_phase(phase, dx = 1, dy = 1, defocus = 0, wavelength = 1.97e-12, divangle = 1e-5):
	r"""Calculate the Lorentz TEM image given a two-dimensional phase distribution and defocus.

	This is an implementation of [2], Eqn (7), which gives the output intensity in terms of \(\phi_m\)
	within the paraxial approximation.

	**Parameters**

	* **phase** : _ndarray_ <br />
	A 2d array containing the phase of the electron at the sample plane.

	* **dx** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the x-direction. <br />
	Default is `dx = 1`.

	* **dy** : _number, optional_ <br />
	The spacing between pixels/samples in mx, my, mz, in the y-direction. <br />
	Default is `dx = 1`.

	* **defocus** : _number, optional_ <br />
	The defocus - note that this should be non-zero in order to see any contrast. <br />
	Default is `defocus = 0`.

	* **wavelength** : _number, optional_ <br />
	The relativistic electron wavelength. <br />
	Default is `wavelength = 1.97e-12`.

	* **divangle** : _number, optional_ <br />
	The divergence angle \(\Theta_c\). <br />
	Default is `divangle = 1e-5`.

	**Returns**

	* _ndarray_ <br />
	The intensity of the image plane.
	"""
	nabla2phase = laplacian_2d(phase, dx, dy)
	nablaphase = gradient_2d(phase, dx, dy)
	nablaphase2 = nablaphase[0]**2 + nablaphase[1]**2

	img = 1 - wavelength * defocus / 2 / _.pi * nabla2phase - (_.pi * divangle * defocus)**2 / 2 / np.log(2) * nablaphase2
	return(img.real)

def propagate(mode, dx = 1, dy = 1, T=T, **kwargs):
	r"""Calculates the Lorentz image given the exit wave \(\psi_0\) and microscope transfer function \(T(\mathbf{q}_{\perp})\).

	\[\psi_f = \mathcal{F}^{-1}\left[\mathcal{F}\left[\psi_0\right] T(\mathbf{q}_{\perp}) \right]\]

	**Parameters**

	* **mode**: _complex ndarray_ <br />
	The exit wave to propagate. Dimension should be 2, may be complex or real.

	* **dx**: _number, optional_ <br />
	The x pixel size of the exit wave. <br />
	Default is `dx = 1`.

	* **dy**: _number, optional_ <br />
	The y pixel size of the exit wave. <br />
	Default is `dx = 1`.

	* **T**: _function, optional_ <br />
	The microscope transfer function. Takes **qx** and **qy** (the spatial frequencies)
	as the first two positional arguments. <br />
	The default is a common transfer function described in Ref (1), Eqns (4-6), that takes the following kwargs: <br />
		<ul>
			<li> **defocus** : _number, optional_ <br />
			Default is `defocus = 1e-3`.
			</li>
			<li> **wavelength** : _number, optional_ <br />
			Default is `wavelength = 1.97e-12` (for 300keV electrons).
			</li>
			<li> **C_s** : _number, optional_ <br />
			The spherical aberration coefficient of the microscope. <br />
			Default is `C_s = 2.7e-3`.
			</li>
			<li> **divangle** : _number, optional_ <br />
			The divergence angle. <br />
			Default is `divangle = 1e-5`.
		</ul>

	* ****kwargs**: _optional_ <br />
	Extra arguments to be passed to the transfer function.

	**Returns**

	* _complex ndarray_ <br />
	The transverse complex amplitude in the image plane. Output has the same
	shape as x, y, and mode.
	"""
	U = np.fft.fftfreq(mode.shape[1], dx)
	V = np.fft.fftfreq(mode.shape[0], dy)
	qx, qy = np.meshgrid(U, V)
	psi_q = np.fft.fft2(mode)
	psi_out = np.fft.ifft2(psi_q * T(qx, qy, **kwargs))
	return(psi_out)

# Miscellaneous
def jchessmodel(x, y, z=0, **kwargs):
	"""Calculates the magnetization of a hopfion based on Jordan Chess' model.

	**Parameters**

	* **x** : _number, ndarray_ <br />
	The x-coordinates over which to calculate magnetization.

	* **y** : _number, ndarray_ <br />
	The y-coordinates over which to calculate magnetization.

	* **z** : _number, ndarray, optional_ <br />
	The z-coordinates over which to calculate magnetization. Note, if z is an
	ndarray, then x, y, and z should have the same shape rather than relying
	on array broadcasting. <br />
	Default is `z = 0`.

	* **aa**, **ba**, **ca** : _number, optional_ <br />
	In this model, the thickness of the domain wall is set by a
	Gaussian function, defined as `aa * exp(-ba * z**2) + ca`. <br />
	Defaults are `aa = 5`, `ba = 5`, `ca = 0`.

	* **ak**, **bk**, **ck** : _number, optional_ <br />
	In this model, the thickness of the core is set by a Gaussian function,
	defined as `ak * exp(-bk * z**2) + ck`. <br />
	Defaults are `ak = 5e7`, `bk = -50`, `ck = 0`.

	* **bg**, **cg** : _number, optional_ <br />
	In this model, the helicity varies as a function of z, given
	as `pi / 2 * tanh( bg * z ) + cg`. <br />
	Defaults are `bg = 5e7`, `cg = pi/2`.

	* **n** : _number, optional_ <br />
	The skyrmion number. <br />
	Default is `n = 1`.

	**Returns**

	* _ndarray_ <br />
	The x-component of magnetization. Shape will be the same as x and y.

	* _ndarray_ <br />
	The y-component of magnetization. Shape will be the same as x and y.

	* _ndarray_ <br />
	The z-component of magnetization. Shape will be the same as x and y.
	"""
	p = {   'aa':5, 'ba':5, 'ca':0,
			'ak':5e7, 'bk':-5e1, 'ck':0,
			'bg':5e7, 'cg':_.pi/2, 'n': 1}
	for key in kwargs.keys():
		if not key in p.keys(): return("Error: {:} is not a kwarg.".format(key))
	p.update(kwargs)

	r, phi = np.sqrt(x**2+y**2), np.arctan2(y,x)

	alpha_z = p['aa'] * np.exp(-p['ba'] * z**2) + p['ca']
	k_z = p['ak'] * np.exp(-p['bk'] * z**2) + p['ck']
	gamma_z = _.pi / 2 * np.tanh(p['bg'] * z) + p['cg']
	Theta_rz = 2 * np.arctan2((k_z * r)**alpha_z, 1)

	mx = np.cos(p['n']*phi%(2*_.pi)-gamma_z) * np.sin(Theta_rz)
	my = np.sin(p['n']*phi%(2*_.pi)-gamma_z) * np.sin(Theta_rz)
	mz = np.cos(Theta_rz)
	return(np.array([mx, my, mz]))



# ltempy is a set of LTEM analysis and simulation tools developed by WSP as a member of the McMorran Lab
# Copyright (C) 2021  William S. Parker
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from . import constants as _


def weights(M, s, s_mag, sig, z_hat, p, thickness):
    # Calculates the weights for the Mansuripur algo
    # aka, the terms in the sum on the RHS of Eq 13(a), without exp(2\pi i ...)
    Gp = G(p, sig, z_hat, thickness * s_mag)
    sig_x_z = np.cross(sig, z_hat, axisa=0, axisb=0, axisc=0)
    p_x_p_M = np.cross(p, np.cross(p, M, axisa=0, axisb=0,
                       axisc=0), axis=0, axisb=0, axisc=0)
    weights = 2 * _.e / _.hbar / _.c * 1j * thickness / s_mag * \
        Gp * np.einsum('i...,i...->...', sig_x_z, p_x_p_M)
    weights[:, 0, 0, :] = 0
    return(weights)


def laplacian_2d(data, dx, dy):
    # 2-dimensional laplacian, implemented via Fourier transform
    QX = np.fft.fftfreq(data.shape[1], dx)
    QY = np.fft.fftfreq(data.shape[0], dy)
    qx, qy = np.meshgrid(QX, QY)
    out = - 4 * _.pi**2 * (qx**2 + qy**2) * np.fft.fft2(data)
    out = np.fft.ifft2(out)
    return(out)


def inverse_laplacian_2d(data, dx, dy):
    # 2-dimensional inverse laplacian, implemented via Fourier transform
    QX = np.fft.fftfreq(data.shape[1], dx)
    QY = np.fft.fftfreq(data.shape[0], dy)
    qx, qy = np.meshgrid(QX, QY)
    out = np.nan_to_num(- np.fft.fft2(data) / 4 / _.pi **
                        2 / (qx**2 + qy**2), posinf=0, neginf=0)
    out = np.fft.ifft2(out)
    return(out)


def gradient_2d(data, dx, dy):
    # 2-dimensional gradient, implemented via Fourier transform
    QX = np.fft.fftfreq(data.shape[1], dx)
    QY = np.fft.fftfreq(data.shape[0], dy)
    qx, qy = np.meshgrid(QX, QY)
    out = np.fft.fft2(data)
    out_x = - 1j * 2 * _.pi * qx * out
    out_y = - 1j * 2 * _.pi * qy * out
    out_x = np.fft.ifft2(out_x)
    out_y = np.fft.ifft2(out_y)
    return(np.array([out_x, out_y]))


def T(qx, qy, defocus=1e-3, wavelength=1.97e-12, C_s=2.7e-3, divangle=1e-5):
    """Utility function for propagate(). Microscope transfer function.
    """
    out = aperture(qx, qy) * np.exp(-1j * chi(qx, qy, defocus, wavelength, C_s)
                                    ) * np.exp(-damping(qx, qy, defocus, wavelength, C_s, divangle))
    return(out)


def damping(qx, qy, defocus=1e-3, wavelength=1.97e-12, C_s=2.7e-3, divangle=1e-5):
    # Damping function for microscope transfer fct
    qp = np.sqrt(qx**2 + qy**2)
    return((_.pi * divangle / wavelength)**2 * (C_s * wavelength**3 * qp**3 + defocus * wavelength * qp)**2)


def chi(qx, qy, defocus, wavelength, C_s):
    """Utility function for propagate(). Phase transfer function.
    """
    return(_.pi * wavelength * defocus * (qx**2 + qy**2) + 0.5 * _.pi * C_s * wavelength**3 * (qx**2 + qy**2)**2)


def aperture(qx, qy, radius=None):
    """Utility function for propagate(). Circular aperture.
    """
    if radius is None:
        radius = np.min([np.max(np.abs(qx)), np.max(np.abs(qy))])
    out = np.zeros_like(qx)
    out[np.sqrt(qx**2 + qy**2) <= radius] = 1
    return(out)


def G(p, sig, z_hat, ts_mag):
    # Mansuripur, Eq 13(b)
    sum1 = np.einsum('i,i...->...', p, sig)
    sum2 = np.einsum('i,i...->...', p, z_hat)
    out = 1 / (sum1**2 + sum2**2)[np.newaxis, ...]
    out = out * np.sinc(ts_mag * sum1 / sum2)
    return(out)


def sims_shared(mx, my, mz, dx, dy):
    # Used for everything in LTEM sims
    # All outputs have shape (3, y-dim, x-dim, z-dim)
    # They just need these four dimensions so that they're broadcastable
    # mx, my, mz either 2 or 3 dim

    mx = np.atleast_3d(mx)  # (y-dim, x-dim, z-dim)
    my = np.atleast_3d(my)
    mz = np.atleast_3d(mz)
    Mx = 1 / mx.shape[0] / mx.shape[1] * np.fft.fft2(mx, axes=(0, 1))
    My = 1 / my.shape[0] / my.shape[1] * np.fft.fft2(my, axes=(0, 1))
    Mz = 1 / mz.shape[0] / mz.shape[1] * np.fft.fft2(mz, axes=(0, 1))
    M = np.array([Mx, My, Mz])  # (vec, y-dim, x-dim, z-dim)

    Sx = np.fft.fftfreq(mx.shape[1], dx)
    Sy = np.fft.fftfreq(mx.shape[0], dy)
    sx, sy = np.meshgrid(Sx, Sy)  # (y-dim, x-dim)

    # (vec, y-dim, x-dim, z-dim)
    s = np.array([sx, sy, 0 * sy])[..., np.newaxis]
    s_mag = np.sqrt(np.einsum('i...,i...->...', s, s)
                    )[np.newaxis, ...]  # (vec, y-dim, x-dim, z-dim)
    sig = s/s_mag

    # (vec, y-dim, x-dim, z-dim)
    z_hat = np.array([np.zeros_like(mx), np.zeros_like(mx), np.ones_like(mx)])
    return(M, s, s_mag, sig, z_hat)


def A_mn_components(
        xshape, yshape, zshape, selz_m, selz_z,
        selz_p, s_mag, z, sigm, sigp, sig, M, thickness, z_hat):
    # set everything to be broadcastable so we can write the A_mn equations
    # shape is: (3 vector components, y-axis, x-axis, z-axis) (thank meshgrid for switching x and y)

    A_mn = np.zeros((3, xshape, yshape, zshape), dtype=complex)
    A_mn[..., selz_m] = (2 * 1j / s_mag
                         * np.exp(2 * _.pi * s_mag * z[..., selz_m])
                         * np.sinh(_.pi * thickness * s_mag)
                         * np.cross(sigm, M, axisa=0, axisb=0, axisc=0))
    A_mn[..., selz_z] = (2 * 1j / s_mag * np.cross((
        sig
        - 0.5 * np.exp(2 * _.pi * s_mag *
                       (z[..., selz_z] - thickness / 2)) * sigm
        - 0.5 * np.exp(-2 * _.pi * s_mag *
                       (z[..., selz_z] + thickness / 2)) * sigp
    ), M, axisa=0, axisb=0, axisc=0))
    A_mn[..., selz_p] = (2 * 1j / s_mag
                         * np.exp(-2 * _.pi * s_mag * z[..., selz_p])
                         * np.sinh(_.pi * thickness * s_mag)
                         * np.cross(sigp, M, axisa=0, axisb=0, axisc=0)
                         )
    zero_comp = np.cross(
        z_hat[:, 0, 0, :], M[:, 0, 0, :], axisa=0, axisb=0, axisc=0)
    A_mn[:, 0, 0, selz_z] = -4 * _.pi * z[:, 0, 0, selz_z] * zero_comp
    A_mn[:, 0, 0, selz_m] = 2 * _.pi * thickness * zero_comp
    A_mn[:, 0, 0, selz_p] = -2 * _.pi * thickness * zero_comp
    return(A_mn)


def _extend_and_fill_mirror(data):
    ds0 = data.shape[0]
    ds1 = data.shape[1]
    bdata = np.zeros((3 * ds0, 3 * ds1), dtype=data.dtype)
    bdata[0:ds0, 0:ds1] = data[::-1, ::-1]
    bdata[0:ds0, ds1:2*ds1] = data[::-1, :]
    bdata[0:ds0, 2*ds1:3*ds1] = data[::-1, ::-1]
    bdata[ds0:2*ds0, 0:ds1] = data[:, ::-1]
    bdata[ds0:2*ds0, ds1:2*ds1] = data
    bdata[ds0:2*ds0, 2*ds1:3*ds1] = data[:, ::-1]
    bdata[2*ds0:3*ds0, 0:ds1] = data[::-1, ::-1]
    bdata[2*ds0:3*ds0, ds1:2*ds1] = data[::-1]
    bdata[2*ds0:3*ds0, 2*ds1:3*ds1] = data[::-1, ::-1]
    return(bdata)



from setuptools import setup, find_packages

setup(
	name = 'ltempy',
	packages = find_packages(),
	version = '1.3.6',
	author = 'William S. Parker',
	author_email = 'will.parker0@gmail.com',
	description = 'Utilities for Lorentz TEM data analysis and simulation.',
	url = 'https://github.com/McMorranLab/ltempy',
	project_urls={
		"Documentation" : "https://mcmorranlab.github.io/ltempy/",
		"Bug Tracker": "https://github.com/McMorranLab/ltempy/issues",
	},
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
		"Operating System :: OS Independent",
	],
	long_description = open('README.md').read(),
	long_description_content_type = "text/markdown",
	python_requires='>=3.6',
	install_requires=['numpy', 'matplotlib']
)



# ltempy is a set of LTEM analysis and simulation tools developed by WSP as a member of the McMorran Lab
# Copyright (C) 2021  William S. Parker
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
`ltempy` is a set of tools for the analysis, simulation, and presentation of Lorentz TEM data.

Features:

* Single Image Transport of Intensity Equation (SITIE) phase reconstruction
* simulations of phase, magnetic field, magnetic vector potential, and Lorentz images
* basic image processing
* a `matplotlib.pyplot` wrapper tailored to presenting induction maps and Lorentz data
* an implementation of the CIELAB perceptually-uniform colorspace
* module-wide unit scaling (i.e., working in nm rather than m)
"""

import numpy as np
import os
from pathlib import Path

from .colors import *
from .constants import set_units
from .process import *
from .simulate import *
from .sitie import *
from .plot import *



# ltempy is a set of LTEM analysis and simulation tools developed by WSP as a member of the McMorran Lab
# Copyright (C) 2021  William S. Parker
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

r"""Contains wrappers for `matplotlib.pyplot`, tailored to the presentation of LTEM data.

The primary feature is the `singleAx` object, which extends the `maplotlib.axes.Axes` object.
It adds methods for windowing data, quiver plots, CIELAB plots, and creating square insets and a colorwheel.

The typical use case is to use `subplots` to generate a figure and axes:

```python
import numpy
import ltempy
# Generate data
X = numpy.linspace(-1, 1, 128)
Y = numpy.linspace(-3, 3, 3 * 128)
x, y = numpy.meshgrid(X, Y)
z = x + 1j*y
f = numpy.sin(z)

# Plot data
window = (.3, .7, .3, .7)
fig, [[ax1, ax2]] = ltempy.subplots(12)
ax1.set_axes(X, Y)
ax1.inset(window)
ax1.imshow(numpy.abs(f)**2)
ax2.set_axes(ax1.x, ax1.y)
ax2.set_window(window)
ax2.cielab(f)
ax2.quiver(f, step=4)
plt.show()
```
"""

import numpy as np
from .colors import cielab_rgba, rgba, cielab_cmap
from . import constants as _

import matplotlib.pyplot as plt


__all__ = ['singleAx', 'subplots']

class singleAx():
  """An extension of the `matplotlib.axes.Axes` class.

  This class adds macros for 2d plotting. In particular,
  it's easy to select only a window of your data to show, to add x-y axes,
  to add an inset, to show the rgba version of a complex 2d array, and to
  show a quiver plot of a complex 2d array.

  Typical usage:

  ```python
  X = np.linspace(-10,10,xres)
  Y = np.linspace(-10,10,yres)
  x, y = np.meshgrid(X, Y)
  data = x+1j*y
  window = [-3,7,1,4]

  fig, ax = plt.subplots()
  myax = ltempy.singleAx(ax)
  myax.set_axes(x, y)
  myax.set_window(window)
  myax.set_xytitle('x','y','title')
  myax.cielab(data)
  plt.show()
  ```

  You also have direct access to the `matplotlib.pyplot.axes.Axes` object via `myax.ax`.

  More commonly, this class is returned by `ltempy.plot.subplots`.

  **Parameters**

  * **ax** : _matplotlib.axes.Axes_ <br />

  **Returns**

  * _ltempy.plot.singleAx_ <br />
  """
  def __init__(self, ax):
    self.ax = ax
    self.origin = 'lower'
    self.shift = 0
    self.x = None
    self.y = None
    self.xmin = None
    self.ymin = None
    self.xmax = None
    self.ymax = None

  def set_title(self, title='', **kwargs):
    """Sets the title of the plot.

    **Parameters**

    * **title** : _string, optional_ <br />
    The plot title. <br />
    Default is `title = ""`.

    * ****kwargs** <br />
    All other kwargs are passed on to `matplotlib.axes.Axes.set_title`.

    **Returns**

    * _singleAx_
    """
    self.ax.set_title(title, **kwargs)
    return(self)

  def set_xlabel(self, xlabel='', **kwargs):
    """Sets the xlabel of the plot.

    **Parameters*

    * **xlabel** : _string, optional_ <br />
    The xlabel. <br />
    Default is `xlabel = ""`.

    * ****kwargs** <br />
    All other kwargs are passed on to `matplotlib.axes.Axes.set_xlabel`.

    **Returns**

    * _singleAx_
    """
    self.ax.set_xlabel(xlabel, **kwargs)
    return(self)

  def set_ylabel(self, ylabel='', **kwargs):
    """Sets the ylabel of the plot.

    **Parameters**

    * **ylabel** : _string, optional_ <br />
    The ylabel. <br />
    Default is `ylabel = ""`.

    * ****kwargs** <br />
    All other kwargs are passed on to `matplotlib.axes.Axes.set_ylabel`.

    **Returns**

    * _singleAx_
    """
    self.ax.set_ylabel(ylabel, **kwargs)
    return(self)

  def set_xytitle(self, xlabel='', ylabel='', title='', **kwargs):
    """Set the xlabel, ylabel, and title at the same time.

    Sets all three even if not all are given.

    For individual control, use `singleAx.set_xlabel`, `singleAx.set_ylabel`,
    or `singleAx.set_title`.

    **Parameters**

    * **ylabel** : _string, optional_ <br />
    The ylabel. <br />
    Default is `ylabel = ""`.

    * **xlabel** : _string, optional_ <br />
    The xlabel. <br />
    Default is `xlabel = ""`.

    * **title** : _string, optional_ <br />
    The plot title. <br />
    Default is `title = ""`.

    * ****kwargs** <br />
    All other kwargs are passed on
    to `matplotlib.axes.Axes.set_xlabel`, `matplotlib.axes.Axes.set_ylabel`,
    and `matplotlib.axes.Axes.set_title`.

    **Returns**

    * _singleAx_
    """
    self.ax.set_xlabel(xlabel, **kwargs)
    self.ax.set_ylabel(ylabel, **kwargs)
    self.ax.set_title(title, **kwargs)
    return(self)

  def set_axes(self, x, y):
    """Sets the x and y axes of the singleAx object.

    Note that this can be used before or after `set_window()`,
    but make sure the two are in the same units.

    **Parameters**

    * **x** : _ndarray_ <br />
    The x-coordinates. Should be 1-dimensional.

    * **y** : _ndarray_ <br />
    The y-coordinates. Should be 1-dimensional.

    **Returns**

    * _singleAx_
    """
    self.x = x
    self.y = y
    return(self)

  def set_window(self, window):
    """Applies a window to the 'singleAx' object.

    Note that this can be used before or after `set_axes()`,
    but make sure the two are in the same units.

    **Parameters**

    * **window** : _array-like_ <br />
    Format: `window = [xmin, xmax, ymin, ymax]`. Note that these are the x
    and y values, rather than their indices.

    **Returns**

    * _singleAx_
    """
    self.xmin = window[0]
    self.xmax = window[1]
    self.ymin = window[2]
    self.ymax = window[3]
    return(self)

  def _pre_plot(self, data, step=1):
    if self.x is None:
      self.x = np.linspace(0, 100, data.shape[1])
    if self.y is None:
      self.y = np.linspace(0, 100, data.shape[0])
    if self.xmin is None:
      self.xmin = self.x[0]
    if self.xmax is None:
      self.xmax = self.x[-1]
    if self.ymin is None:
      self.ymin = self.y[0]
    if self.ymax is None:
      self.ymax = self.y[-1]
    argxmin = np.argmin(np.abs(self.x - self.xmin))
    argxmax = np.argmin(np.abs(self.x - self.xmax))
    argymin = np.argmin(np.abs(self.y - self.ymin))
    argymax = np.argmin(np.abs(self.y - self.ymax))
    dout = data[argymin:argymax:step, argxmin:argxmax:step]
    xout = self.x[argxmin:argxmax:step]
    yout = self.y[argymin:argymax:step]
    return(xout, yout, dout)

  def imshow(self, data, step=1, colorbar=False, **kwargs):
    """Imshows the (windowed) data.

    **Parameters**

    * **data** : _ndarray_ <br />
    The data to be shown. Use the un-windowed data - the window will be
    applied automatically, if you have set one.

    * **step** : _int, optional_ <br />
    data will be shown as `data[::step,::step]`. <br />
    Default is `step = 1`.

    * **colorbar** : _boolean, optional_ <br />
    If `True`, a colorbar will be added next to the plot. For more control over its
    appearance, see `singleAx.colorbar()`. <br />
    Default is `colorbar = False`.

    * ****kwargs** <br />
    All other kwargs are passed on to `matplotlib.axes.Axes.imshow`.

    **Returns**

    * _matplotlib.AxesImage_ <br />
    Returns the 'AxesImage' returned by `matplotlib.axes.Axes.imshow()`.
    """
    imshowargs = {'origin': self.origin}
    imshowargs.update(kwargs)
    x, y, d = self._pre_plot(data, step)
    if imshowargs['origin'] == 'lower':
      extent = [x[0], x[-1], y[0], y[-1]]
    elif imshowargs['origin'] == 'upper':
      extent = [x[0], x[-1], y[-1], y[0]]
    imshowargs.update({'extent': extent})
    imshowargs.update(kwargs)
    im = self.ax.imshow(d, **imshowargs)
    if colorbar:
      self.colorbar(im)
    return(im)

  def quiver(self, data, step=1, origin=None, **kwargs):
    """Shows a quiver plot of complex data.

    **Parameters**

    * **data** : _ndarray, complex_ <br />
    The data to be shown. Real part is x-component, imaginary is y-component.
    Use the un-windowed data - the window will be applied automatically, if you set one.

    * **step** : _int_ <br />
    data will be shown as `data[::step,::step]`. <br />
    Default is `step = 1`.

    * **origin** : _string_ <br />
    Either 'upper' or 'lower'. <br />
    Default is `self.origin`, whose default is 'lower'.

    * ****kwargs** <br />
    All other kwargs are passed on to `matplotlib.axes.Axes.quiver`.

    **Returns**

    * _matplotlib.quiver.Quiver_ <br />
    Returns the `Quiver` object returned by `matplotlib.axes.Axes.quiver()`.
    """
    if origin is None:
      origin = self.origin
    x, y, d = self._pre_plot(data, step)
    d = d.astype(complex)
    if origin == 'upper':
      d.imag = -1 * d.imag
    return(self.ax.quiver(x, y, d.real, d.imag, **kwargs))

  def cielab(self, data, step=1, brightness='intensity', alpha='uniform', **kwargs):
    """Show a CIELAB interpretation of complex data.

    Color represents phase, while brightness may be uniform or represent amplitude or intensity.

    **Parameters**

    * **data** : _complex ndarray_ <br />
    An array with the data to represent. Dtype may be complex or real.

    * **step** : _int_ <br />
    data will be shown as `data[::step,::step]`. <br />
    Default is `step = 1`.

    * **brightness** : _string, optional_ <br />
    Allowed values: `'intensity'`, `'amplitude'`, `'uniform'`. <br />
    Default is `brightness = 'intensity'`.

    * **alpha** : _string, optional_ <br />
    Allowed values: `'intensity'`, `'amplitude'`, `'uniform'`. Determines the alpha
    component of the rgba value. <br />
    Default is `alpha = 'uniform'`.

    * ****kwargs** <br />
    All other kwargs are passed on to `matplotlib.axes.Axes.imshow`.

    **Returns**

    * _matplotlib.AxesImage_ <br />
    Returns the 'AxesImage' returned by `matplotlib.axes.Axes.imshow()`.
    """
    imshowargs = {'origin': self.origin}
    imshowargs.update(kwargs)
    x, y, d = self._pre_plot(data, step)
    d = d.astype(complex)
    if imshowargs['origin'] == 'lower':
      extent = [x[0], x[-1], y[0], y[-1]]
    elif imshowargs['origin'] == 'upper':
      extent = [x[0], x[-1], y[-1], y[0]]
      d.imag = -1 * d.imag
    imshowargs.update({'extent': extent})
    imshowargs.update(kwargs)
    im = self.ax.imshow(rgba(d, brightness=brightness, alpha=alpha, shift=self.shift), **imshowargs)
    return(im)

  def colorbar(self, axesImage, position='right', size='5%', pad=0.05, **kwargs):
    """Append a colorbar to the `axes` object, based on the provided `axesImage`.

    There are two ways to apply a colorbar. You can provided `colorbar = True` to `ax.imshow()`. This will create a colorbar with some default parameters.
    Otherwise, if you want to customize the colorbar, you can use this method to do the following:
    ```python
    fig, [[ax]] = ltempy.subplots()
    im = ax.imshow(data)
    ax.colorbar(im)
    plt.show()
    ```
    and pass extra arguments to `colorbar()`.

    **Parameters**

    * **axesImage** : _matplotlib.image.AxesImage_ <br />
    This is what is returned by, for example, `plt.imshow()`.

    * **position** : _string_ <br />
    Accepted values are `'left'`, `'right'`, `'top'`, `'bottom'`. <br />
    Default is `position = "right"`.

    * **size** : _string_ <br />
    Must be `mpl_toolkits.axes_grid.axes_size` compatible. For example, give a percent. <br />
    Default is `size = "5%"`.

    * **pad** : _number_ <br />
    Default is `pad = 0.05`.

    * ****kwargs** <br />
    All further kwargs are passed to `Figure.colorbar()`.

    **Returns**

    * _matplotlib.colorbar.colorbar_
    """
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    divider = make_axes_locatable(self.ax)
    cax = divider.append_axes(position, size, pad)
    fig = self.ax.get_figure()
    cbar = fig.colorbar(axesImage, cax=cax, **kwargs)
    return(cbar)

  def inset(self, window, **kwargs):
    """Plots a square box with vertices defined by window.

    Default color is white.

    **Parameters**

    * **window** : _array-like_ <br />
    Format: `window = [xmin, xmax, ymin, ymax]`. Note that these are the x
    and y values, rather than their indices.

    * ****kwargs** <br />
    All other kwargs are passed on to `matplotlib.axes.Axes.plot`.

    **Returns**

    * _singleAx_
    """
    plotargs = {'color': 'white', 'linewidth': .5}
    plotargs.update(kwargs)
    self.ax.plot(np.linspace(window[0], window[1], 100),
            np.zeros(100) + window[2], **plotargs)
    self.ax.plot(np.linspace(window[0], window[1],100),
            np.zeros(100)+window[3], **plotargs)
    self.ax.plot(np.zeros(100) + window[0],
            np.linspace(window[2], window[3], 100), **plotargs)
    self.ax.plot(np.zeros(100) + window[1],
            np.linspace(window[2], window[3], 100),
            **plotargs)
    return(self)

  def colorwheel(self, res=128, scale=0.25, cmap=None, brightness='intensity', alpha='uniform', **kwargs):
    """Adds a colorwheel to the bottom right corner of the plot.

    **Parameters**

    * **res** : _int_ <br />
    The resolution of the colorwheel. <br />
    Default is `res = 128`.

    * **scale** : _float_ <br />
    The size of the colorwheel, in units of the width of the axis. <br />
    Default is `scale = 0.25`.

    * **cmap** : _string, optional_ <br />
    If `cmap = None`, the CIELAB color space will be used. Otherwise, any
    pyplot ScalarMappable may be used. <br />
    Default is `cmap = None`.

    * **brightness** : _string, optional_ <br />
    Allowed values: `'intensity'`, `'amplitude'`, `'uniform'`. <br />
    Default is `brightness = 'intensity'`.

    * **alpha** : _string, optional_ <br />
    Allowed values: `'intensity'`, `'amplitude'`, `'uniform'`. Determines the alpha
    component of the rgba value. <br />
    Default is `alpha = 'uniform'`.

    * ****kwargs** <br />
    All other kwargs are passed on to `matplotlib.axes.Axes.imshow`.

    **Returns**

    * _singleAx_
    """
    imshowargs = {'origin': self.origin, 'zorder': 3}
    imshowargs.update(kwargs)
    X = np.linspace(-1, 1, res)
    x, y = np.meshgrid(X, X)
    z = x + 1j*y
    sel = np.abs(z) > 1
    z[sel] = 0
    colors = rgba(z, brightness=brightness, alpha=alpha, cmap=cmap, shift=self.shift)
    colors[:,:,0][sel] = 0
    colors[:,:,1][sel] = 0
    colors[:,:,2][sel] = 0
    xlims = self.ax.get_xlim()
    ylims = self.ax.get_ylim()
    self.ax.set_xlim(xlims)
    self.ax.set_ylim(ylims)
    self.ax.set_aspect('equal')
    if self.origin == 'lower':
      extent = [xlims[1] - scale * (xlims[1] - xlims[0]),
            xlims[1],
            ylims[0],
            ylims[0] + scale * (xlims[1] - xlims[0])]
    elif self.origin == 'upper':
      extent = [xlims[1] - scale * (xlims[1] - xlims[0]),
            xlims[1],
            ylims[0] + scale * (xlims[0] - xlims[1]),
            ylims[0]]
    imshowargs.update(kwargs)
    self.ax.imshow(colors, extent=extent, **imshowargs)
    return(self)

def subplots(rc=11, **kwargs):
    """Creates a figure and axes, where each axis is a `singleAx` instance.

    Behaves like `matplotlib.pyplot.subplots()`, but replaces each
    `matplotlib.axes.Axes` object with a `wsp_tools.plot.singleAx`
    object.

    Each `wsp_tools.plot.singleAx` object in turn behaves like a
    normal `Axes` object, but with added methods.

    Note: two default kwargs are set. `'tight_layout': True` and `'squeeze': False`.

    **Parameters**

    * **rc** : _int_ <br />
    First digit - nrows. Second digit - ncols. <br />
    Default is `rc = 11`.

    * ****kwargs** <br />
    All other kwargs are passed on to `matplotlib.axes.Axes.subplots`.

    **Returns**

    * _Figure_ <br />

    * _singleAx_ or array of _singleAx_ objects <br />
    """
    subplotsargs = {'tight_layout': True, 'squeeze': False, 'nrows':rc//10, 'ncols':rc%10}
    subplotsargs.update(kwargs)
    fig, ax = plt.subplots(**subplotsargs)
    for i in range(ax.shape[0]):
      for j in range(ax.shape[1]):
        ax[i][j] = singleAx(ax[i][j])
    return(fig, np.array(ax))
